Require RAMs.
Require Linda.
Require PolyList.
Require TheoryList_compl.

Definition program_of_instruction :=
 [registers_no: nat ; pc: nat ; i:RAMs.instruction]
  Cases i of
     (Succ r) =>
       (Action (BangIn (plus registers_no pc))
         (Action (Out r)
          (Action (Out (plus registers_no (S pc))) Dead)))
   | (DecJump r n) =>
       (Action (BangIn (plus registers_no pc))
         (IfThenElseAction (Linda.In r)
           (Action (Out (plus registers_no (S pc))) Dead)
           (Action (Out (plus registers_no n)) Dead)))
  end.

Definition agent_of_program :=
 [registers_no: nat ;p : (list RAMs.instruction)]
  (TheoryList_compl.fold_left_i ? ?
    [pc: nat; a: Linda.agent; i: RAMs.instruction]
     (AParallel a (Program (program_of_instruction registers_no pc i)))
    p
    (0)
    (Program Dead)).

Fixpoint agent_of_register [r: register ; v:nat] : agent :=
 Cases v of
    O => (Program Dead)
  | (S n) => (AParallel (Message r) (agent_of_register r n))
 end.

Fixpoint agent_of_registers [f:register -> nat ; n:nat] : agent :=
 Cases n of
    O => (Program Dead)
  | (S m) => (AParallel (agent_of_register m (f m)) (agent_of_registers f m))
 end.

(* A technical recursion principle *)
Theorem agent_of_registers_ind :
 (f:(register->nat) ; registers_no:nat ; P: agent -> Prop)
  (P (Program Dead)) ->
  ((r:nat)
    (lt r registers_no) ->
     (P (agent_of_registers f r)) ->
      (P (AParallel (agent_of_register r (f r)) (agent_of_registers f r)))) ->
  (P (agent_of_registers f registers_no)).
Proof.
 Intros;
 Induction registers_no;
  [ Simpl;
    Exact H
  | Simpl;
    Apply H0;
     [ Auto with arith
     | Apply Hrecregisters_no;
       Intros;
       Apply H0;
        [ Auto with arith
        | Assumption
        ]
     ]
  ].
Qed.

Definition agent_of_configuration :=
 [program_length : nat ; registers_no : nat ; c : RAMs.configuration]
  Cases c of
   (Build_configuration pc registers) =>
     (AParallel
       (Message (plus registers_no pc))
       (agent_of_registers registers registers_no))
  end.

Definition agent_of_RAM_and_configuration :=
 [r: RAM ; c : configuration]
  Cases r of
   (Build_RAM registers_no program _) =>
     (AParallel
       (agent_of_program registers_no program)
       (agent_of_configuration (length program) registers_no c))
  end.
